/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package parameters

import (
	"context"
	"errors"
	"fmt"
	"reflect"
	"strings"

	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime"

	"k8s.libre.sh/interfaces"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/client/apiutil"
)

func MergeParametersWithData(params Parameters, data map[string]string) Parameters {

	paramsByKey := ParametersByKey(&params)

	if len(data) > 0 {
		if len(params) > 0 {
			for _, pp := range params {
				paramsByKey[pp.Key].Value = string(data[pp.FromKey])
				// Reset default Generate
				paramsByKey[pp.Key].Generate = ""
			}
		} else {
			for kk, vv := range data {
				if paramsByKey[kk] == nil {
					paramsByKey[kk] = &Parameter{}

				}
				paramsByKey[kk].Value = vv
				paramsByKey[kk].Key = kk

				// TODO TO FIX This should be checked
				//	paramsByKey[kk].Type = pType

				// Reset default Generate
				paramsByKey[kk].Generate = ""
			}
		}
	}

	// TODO TO FIX
	ps := Parameters{}

	for _, v := range paramsByKey {
		ps = append(ps, v)
	}

	return ps

}

func MergeParameters(src, dest *Parameters) *Parameters {

	params := Parameters{}

	srcByKey := ParametersByKey(src)
	destByKey := ParametersByKey(dest)

	for k, param := range destByKey {
		if srcByKey[k] != nil {
			param = srcByKey[k]
		}
		params = append(params, param)
	}

	return &params
}

func Marshal(i interface{}) (Parameters, error) {
	parameters := Parameters{}

	t := reflect.TypeOf(i)

	ifv := reflect.ValueOf(i)

	for ii := 0; ii < t.NumField(); ii++ {
		f := t.Field(ii)

		fv := ifv.Field(ii)

		parameter, ok := fv.Interface().(Parameter)
		if !ok {
			return parameters, errors.New("not parameters type")
		}

		env, ok := f.Tag.Lookup("env")
		if !ok {
			return parameters, errors.New("missing evnv tag")
		}

		if len(parameter.Key) == 0 {
			parameter.Key = env
		}

		parameters = append(parameters, &parameter)
	}

	return parameters, nil
}

// Sort parameters by keys
func ParametersByKey(p *Parameters) map[string]*Parameter {
	paramsByKey := map[string]*Parameter{}

	for _, param := range *p {
		paramsByKey[param.Key] = param
	}
	return paramsByKey
}

// Transform parameters to key:value
func KeyPairValues(p *Parameters) map[string]string {
	data := make(map[string]string)

	for _, param := range *p {
		data[param.Key] = param.Value
	}

	return data
}

func ToOwnerReference(object metav1.Object, scheme *runtime.Scheme) (ref metav1.OwnerReference, err error) {
	// Validate the owner.
	ro, ok := object.(runtime.Object)
	if !ok {
		return ref, fmt.Errorf("%T is not a runtime.Object, cannot call SetOwnerReference", object)
	}

	// Create a new owner ref.
	gvk, err := apiutil.GVKForObject(ro, scheme)
	if err != nil {
		return ref, err
	}
	ref = metav1.OwnerReference{
		APIVersion: gvk.GroupVersion().String(),
		Kind:       gvk.Kind,
		UID:        object.GetUID(),
		Name:       object.GetName(),
	}

	return ref, nil

}

func GetDataValueFrom(r interfaces.Reconcile, obj, owner interfaces.Object) (data map[string]string, err error) {

	objData := make(map[string]string)
	data = make(map[string]string)

	objectKey, _ := client.ObjectKeyFromObject(obj)

	err = r.GetClient().Get(context.Background(), objectKey, obj)
	if err != nil {
		return data, err
	}

	ownerRef, err := ToOwnerReference(owner, r.GetScheme())
	if err != nil {
		return data, err
	}

	switch obj.GetObjectKind().GroupVersionKind().Kind {
	case "Secret":
		for k, v := range obj.(*corev1.Secret).Data {
			objData[k] = string(v)
		}
	case "ConfigMap":
		objData = obj.(*corev1.ConfigMap).Data
	default:
		return data, errors.New("object kind should be ConfigMap or Secret")
	}

	objOwnerRefs := obj.GetOwnerReferences()
	if len(objOwnerRefs) > 0 {
		for _, ref := range objOwnerRefs {
			if ref.APIVersion == ownerRef.APIVersion && ref.UID == ownerRef.UID && ref.Kind == ownerRef.Kind {
				if len(obj.GetAnnotations()["settings.k8s.libre.sh/generate"]) > 0 {
					genParams := strings.Split(obj.GetAnnotations()["settings.k8s.libre.sh/generate"], ",")
					for _, param := range genParams {
						data[param] = objData[param]
					}
				}
			} else {
				data = objData
			}
		}
	} else {
		data = objData
	}

	return data, nil
}

/*
// TODO SPLIT IN GETPARAMETERSFROMSOURCES AND MERGEPARAMETERSBYKEYS
func GetAndMergeParameters(
	params Parameters,
	paramsByKey map[string]*Parameter,
	r interfaces.Reconcile,
	obj, owner interfaces.Object) error {

	var pType ParameterType

	data, err := GetDataValueFrom(r, obj, owner)
	if err != nil {
		return err
	}

	switch obj.GetObjectKind().GroupVersionKind().Kind {
	case "Secret":
		pType = SecretParameter
	case "ConfigMap":
		pType = ConfigParameter
	default:
		return errors.New("object kind should be ConfigMap or Secret")
	}

	 	data := map[string]string{}

	   	objectKey, _ := client.ObjectKeyFromObject(obj)

	   	err := r.GetClient().Get(context.Background(), objectKey, obj)
	   	if err != nil {
	   		return err
	   	}

	   	ownerRef, err := ToOwnerReference(owner, r.GetScheme())
	   	if err != nil {
	   		return err
	   	}

	   	switch obj.GetObjectKind().GroupVersionKind().Kind {
	   	case "Secret":
	   		pType = SecretParameter
	   		objOwnerRefs := obj.GetOwnerReferences()
	   		if len(objOwnerRefs) > 0 {
	   			for _, ref := range objOwnerRefs {
	   				if ref.APIVersion == ownerRef.APIVersion && ref.UID == ownerRef.UID && ref.Kind == ownerRef.Kind {
	   					if len(obj.GetAnnotations()["settings.k8s.libre.sh/generate"]) > 0 {
	   						genParams := strings.Split(obj.GetAnnotations()["settings.k8s.libre.sh/generate"], ",")
	   						for _, param := range genParams {
	   							data[param] = string(obj.(*corev1.Secret).Data[param])
	   						}
	   					}
	   				} else {
	   					for k, v := range obj.(*corev1.Secret).Data {
	   						data[k] = string(v)
	   					}
	   				}
	   			}
	   		} else {
	   			for k, v := range obj.(*corev1.Secret).Data {
	   				data[k] = string(v)
	   			}
	   		}

	   	case "ConfigMap":
	   		pType = ConfigParameter
	   		for k, v := range obj.(*corev1.ConfigMap).Data {
	   			data[k] = v
	   		}
	   	default:
	   		return errors.New("object kind should be ConfigMap or Secret")
	   	}

	if len(params) > 0 {
		for _, pp := range params {
			paramsByKey[pp.Key].Value = string(data[pp.FromKey])
			paramsByKey[pp.Key].Generate = ""
		}
	} else {
		for kk, vv := range data {
			if paramsByKey[kk] == nil {
				paramsByKey[kk] = &Parameter{}

			}
			paramsByKey[kk].Value = vv
			paramsByKey[kk].Key = kk
			// TODO TO FIX This should be checked
			paramsByKey[kk].Type = pType
			paramsByKey[kk].Generate = ""
		}
	}

	return nil
}
*/
